#
# This file is part of Invenio.
# Copyright (C) 2016-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Create member tables"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy_utils import UUIDType

# revision identifiers, used by Alembic.
revision = 'fbe746957cfc'
down_revision = 'f701a32e6fbe'
branch_labels = ()
depends_on = None


def upgrade():
    """Upgrade database."""
    op.add_column(
        'communities_members',
        sa.Column('active', sa.Boolean(), nullable=False))
    op.add_column(
        'communities_members',
        sa.Column('visible', sa.Boolean(), nullable=False))
    op.add_column(
        'communities_members',
        sa.Column('group_id', sa.Integer(), nullable=True))
    op.add_column(
        'communities_members',
        sa.Column('request_id', UUIDType(), nullable=True))
    op.create_index(
        op.f('ix_communities_members_active'),
        'communities_members',
        ['active'],
        unique=False
    )
    op.create_index(
        'ix_community_group',
        'communities_members',
        ['community_id', 'group_id'],
        unique=True
    )
    op.create_unique_constraint(
        op.f('uq_communities_members_request_id'),
        'communities_members',
        ['request_id']
    )
    op.drop_constraint(
        'fk_communities_members_user_id_accounts_user',
        'communities_members',
        type_='foreignkey'
    )
    op.create_foreign_key(
        op.f('fk_communities_members_user_id_accounts_user'),
        'communities_members', 'accounts_user', ['user_id'], ['id'],
        ondelete='RESTRICT'
    )
    op.create_foreign_key(
        op.f('fk_communities_members_group_id_accounts_role'),
        'communities_members', 'accounts_role', ['group_id'], ['id'],
        ondelete='RESTRICT'
    )
    op.create_foreign_key(
        op.f('fk_communities_members_request_id_request_metadata'),
        'communities_members', 'request_metadata', ['request_id'], ['id'],
        ondelete='SET NULL'
    )



def downgrade():
    """Downgrade database."""
    op.drop_constraint(op.f('fk_communities_members_request_id_request_metadata'), 'communities_members', type_='foreignkey')
    op.drop_constraint(op.f('fk_communities_members_group_id_accounts_role'), 'communities_members', type_='foreignkey')
    op.drop_constraint(op.f('fk_communities_members_user_id_accounts_user'), 'communities_members', type_='foreignkey')
    op.create_foreign_key('fk_communities_members_user_id_accounts_user', 'communities_members', 'accounts_user', ['user_id'], ['id'], ondelete='CASCADE')
    op.drop_constraint(op.f('uq_communities_members_request_id'), 'communities_members', type_='unique')
    op.drop_index('ix_community_group', table_name='communities_members')
    op.drop_index(op.f('ix_communities_members_active'), table_name='communities_members')
    op.drop_column('communities_members', 'request_id')
    op.drop_column('communities_members', 'group_id')
    op.drop_column('communities_members', 'visible')
    op.drop_column('communities_members', 'active')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Member schema."""

from flask_babelex import lazy_gettext as _
from marshmallow import Schema, ValidationError, fields, pre_dump, validate, \
    validates_schema
from marshmallow_utils.fields import SanitizedUnicode
from marshmallow_utils.fields import TZDateTime
from datetime import timezone

from .fields import RoleField

#
# Utility schemas
#
class MemberEntitySchema(Schema):
    """Represents a single member entity."""

    type = fields.String(
        required=True,
        validate=validate.OneOf(['user', 'group', 'email'])
    )
    id = fields.String(required=True)
    is_current_user = fields.Bool(dump_only=True)


class MembersSchema(Schema):
    members = fields.List(
        fields.Nested(MemberEntitySchema),
        # max is on purpose to limit the max number of additions/changes/
        # removals per request as they all run in a single transaction and
        # requires resources to hold.
        validate=validate.Length(min=1, max=100)
    )


#
# Schemas used for validation
#
class AddBulkSchema(MembersSchema, Schema):
    """Schema used for adding members."""

    role = RoleField(required=True)
    visible = fields.Boolean()


class InviteBulkSchema(AddBulkSchema):
    """Schema used for inviting members."""
    message = SanitizedUnicode()


class UpdateBulkSchema(MembersSchema, Schema):
    """Schema used for updating members."""

    role = RoleField()
    visible = fields.Boolean()

    @validates_schema
    def validate_schema(self, data, **kwargs):
        """Validates that role and/or visible is set."""
        if 'role' not in data and 'visible' not in data:
            raise ValidationError(_("Missing fields 'role' and/or 'visible'"))


class DeleteBulkSchema(MembersSchema):
    """Delete bulk schema."""


#
# Schemas used for dumping a single member
#

class PublicDumpSchema(Schema):
    id = fields.String(required=True)
    member = fields.Method('get_member')

    def get_member(self, obj):
        if obj.user_id:
            return {
                "type": "user",
                "id": str(obj.user_id),
                # TODO: fix all below once indexing is fixed
                # TODO: what about email visibility and public visibility
                "name": f"{obj.user_id}-user-name",
                "description": f"{obj.user_id}-user-description",
                "links": {
                    "avatar": "..."
                }
            }
        elif obj.group_id:
            return {
                "type": "group",
                # TODO: fix me, should be group name
                "id": str(obj.group_id),
                # TODO: fix all below once indexing is fixed
                "name": f"{obj.user_id}-group-name",
                "description": f"{obj.user_id}-group-description",
                "links": {
                    "avatar": "..."
                }
            }


class MemberDumpSchema(PublicDumpSchema):
    role = fields.String()
    visible = fields.Boolean()

    is_current_user = fields.Method('get_current_user')
    permissions = fields.Method('get_permissions')

    created = TZDateTime(timezone=timezone.utc, format='iso')
    updated = TZDateTime(timezone=timezone.utc, format='iso')
    revision_id = fields.Integer()

    def is_self(self, obj):
        if 'is_self' not in self.context:
            current_identity = self.context['identity']
            self.context['is_self'] = (
                obj.user_id is not None and
                current_identity.id is not None and
                str(obj.user_id) == str(current_identity.id)
            )
        return self.context['is_self']

    def get_current_user(self, obj):
        return self.is_self(obj)

    def get_permissions(self, obj):
        return {
            # TODO: is_self and not the last owner
            'can_leave': self.is_self(obj),
            'can_delete': False,
            'can_update_role': False,
            # owners/managers, prior value of visibility, self, prior
            'can_update_visible': self.is_self(obj),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"Members data layer API."

from invenio_accounts.models import Role
from invenio_communities.members.errors import InvalidMemberError
from invenio_db import db
from invenio_records.systemfields import ModelField
from invenio_records_resources.records.api import Record
from invenio_records_resources.records.systemfields import IndexField
from sqlalchemy import or_

from .models import MemberModel


class Member(Record):
    """A member record.

    We are using a record without using the actual JSON document and
    schema validation normally used in a record. The reason for using a record
    is to facilitate the indexing which we need to have an effective search
    over the list of members.
    """

    model_cls = MemberModel

    # Systemfields

    metadata = None

    index = IndexField(
        "communitymembers-members-v1.0.0", search_alias="communitymembers"
    )
    """The ES index used."""

    community_id = ModelField("community_id")
    """The data-layer UUID of the community."""

    user_id = ModelField("user_id")
    """The data-layer id of the user (or None)."""

    group_id = ModelField("group_id")
    """The data-layer id of the user (or None)."""

    request_id = ModelField("request_id")
    """The data-layer id of the user (or None)."""

    role = ModelField("role")
    """The role of the entity."""

    visible = ModelField("visible")
    """Visibility of the membership."""

    active = ModelField("active")
    """Determine if it's an active membership.

    This is used for e.g. invitations where a memberships is created but not
    yet activated.
    """

    @classmethod
    def get_memberships(cls, identity):
        """Get community memberships for a given identity."""
        # TODO: extract group/role ids from identity
        query = cls.model_cls.query_memberships(
            user_id=identity.id, group_ids=[]
        )
        return [(str(comm_id), role) for comm_id, role in query]

    @classmethod
    def get_member_by_request(cls, request_id):
        """Get a membership by request id."""
        assert request_id is not None
        obj = cls.model_cls.query.filter(
            cls.model_cls.request_id==request_id
        ).one()
        return cls(obj.data, model=obj)

    @classmethod
    def get_members(cls, community_id, members=None):
        """Get members of a community."""
        # Collect users and groups we are interested in
        user_ids = []
        group_names = []
        for m in (members or []):
            if m['type'] == 'group':
                group_names.append(m['id'])
            elif m['type'] == 'user':
                user_ids.append(m['id'])
            else:
                raise InvalidMemberError(m)

        # Query
        q = cls.model_cls.query.filter(
            cls.model_cls.community_id == community_id
        )

        # Apply user and group query if applicable
        user_q = cls.model_cls.user_id.in_(user_ids)
        groups_q = cls.model_cls.group_id.in_(
            db.session.query(Role.id).filter(Role.name.in_(group_names))
        )
        if user_ids and group_names:
            q = q.filter(or_(user_q, groups_q))
        elif user_ids:
            q = q.filter(user_q)
        elif group_names:
            q = q.filter(groups_q)

        return [cls(obj.data, model=obj) for obj in q.all()]


    @classmethod
    def has_members(cls, community_id, role=None):
        """Get members of a community."""
        return cls.model_cls.count_members(community_id, role=role)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Northwestern University.
# Copyright (C) 2022 CERN.
#
# Invenio-Communities is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Member Model."""

import uuid

from invenio_accounts.models import Role, User
from invenio_db import db
from invenio_records.models import RecordMetadataBase
from invenio_requests.records.models import RequestMetadata
from sqlalchemy import CheckConstraint, Index
from sqlalchemy_utils.types import UUIDType

from ...communities.records.models import CommunityMetadata


class MemberModel(db.Model, RecordMetadataBase):
    """Member model.

    We restrict deletion of users/groups if they are present in the member
    table, to ensure that we have at least one owner. I.e. users must first
    be removed from memberships before they can be deleted, ensuring that
    another owner is set of a community if they are the sole owner.
    """

    __tablename__ = "communities_members"
    __table_args__ = (
        Index('ix_community_user', 'community_id', 'user_id', unique=True),
        Index('ix_community_group', 'community_id', 'group_id', unique=True),
        # Make sure user or group is set but not both.
        CheckConstraint(
            '(user_id IS NULL AND group_id IS NOT NULL) OR '
            '(user_id IS NOT NULL AND group_id IS NULL)',
            name='user_or_group',
        ),
    )

    id = db.Column(UUIDType, primary_key=True, default=uuid.uuid4)

    community_id = db.Column(
        UUIDType,
        db.ForeignKey(CommunityMetadata.id, ondelete="CASCADE"),
        nullable=False
    )

    role = db.Column(db.String(50), nullable=False)

    active = db.Column(db.Boolean(), index=True, nullable=False)

    visible = db.Column(db.Boolean(), nullable=False)

    user_id = db.Column(
        db.Integer(),
        db.ForeignKey(User.id, ondelete="RESTRICT"),
        nullable=True,
    )

    group_id = db.Column(
        db.Integer(),
        db.ForeignKey(Role.id, ondelete="RESTRICT"),
        nullable=True,
    )

    request_id = db.Column(
        UUIDType,
        db.ForeignKey(RequestMetadata.id, ondelete="SET NULL"),
        nullable=True,
        unique=True,
    )
    """An associated request.

    A request can only be associated with one membership/invitation.
    """

    @classmethod
    def query_memberships(cls, user_id=None, group_ids=None, active=True):
        """Query for (community,role)-pairs."""
        q = db.session.query(
            cls.community_id, cls.role
        ).filter(cls.active==active)

        if user_id:
            q = q.filter(cls.user_id==user_id)
        if group_ids:
            q = q.filter(cls.group_id.in_(group_ids))

        return q.distinct()

    @classmethod
    def count_members(cls, community_id, role=None, active=True):
        """Count number of members."""
        q = cls.query.filter(
            cls.community_id==community_id,
            cls.active==active
        )
        if role is not None:
            q = q.filter(cls.role==role)
        return q.count()

